import numpy as np

from test_utils import change_cwd_to_tests_dir, add_project_root_to_path

change_cwd_to_tests_dir()
add_project_root_to_path()

from cvvideoplayer import TestingFrameReader, create_video_player
from cvvideoplayer.utils.ui_utils import SingleInput, InputType


def test_video_navigation():
    frame_reader = TestingFrameReader(video_len=70)
    video_player = create_video_player(
        video_source=frame_reader,
        start_from_frame=5
    )
    assert np.all(video_player._get_current_frame() == frame_reader.get_frame(5))
    video_player.input_handler.handle_input(SingleInput(InputType.KeyPress, "right"))
    assert np.all(video_player._get_current_frame() == frame_reader.get_frame(6))
    video_player.input_handler.handle_input(SingleInput(InputType.KeyPress, "ctrl+right"))
    assert np.all(video_player._get_current_frame() == frame_reader.get_frame(16))
    video_player.input_handler.handle_input(SingleInput(InputType.KeyPress, "ctrl+right"))
    assert np.all(video_player._get_current_frame() == frame_reader.get_frame(26))
    video_player.input_handler.handle_input(SingleInput(InputType.KeyPress, "ctrl+shift+right"))
    assert np.all(video_player._get_current_frame() == frame_reader.get_frame(69))
    video_player.input_handler.handle_input(SingleInput(InputType.KeyPress, "ctrl+shift+left"))
    assert np.all(video_player._get_current_frame() == frame_reader.get_frame(19))
    video_player.input_handler.handle_input(SingleInput(InputType.KeyPress, "ctrl+left"))
    assert np.all(video_player._get_current_frame() == frame_reader.get_frame(9))
    video_player.input_handler.handle_input(SingleInput(InputType.KeyPress, "ctrl+left"))
    assert np.all(video_player._get_current_frame() == frame_reader.get_frame(0))
    video_player.input_handler.handle_input(SingleInput(InputType.KeyPress, "left"))
    assert np.all(video_player._get_current_frame() == frame_reader.get_frame(0))
    video_player.input_handler.handle_input(SingleInput(InputType.KeyPress, "esc"))
    video_player.__exit__()



import ctypes
from ctypes.wintypes import LPCWSTR, HWND
from pathlib import Path

import cv2
import numpy as np

from ..utils.video_player_utils import KeyFunction
from ..input_management.windows_input_parser import WindowsInputParser
from ..utils.windows_os_utils import set_icon_windows
from .base_video_player import VideoPlayer


class WindowsVideoPlayer(VideoPlayer):
    def __init__(self, **video_player_kwargs):
        ctypes.windll.shcore.SetProcessDpiAwareness(2)
        super().__init__(**video_player_kwargs)
        self._zoom_factor = 1.0
        self._zoom_crop_xywh = None
        current_frame = self._get_current_frame()
        self._original_frame_size = tuple(current_frame.shape[:2][::-1])
        self._zoom_crop_xywh = (
            0,
            0,
            self._original_frame_size[0],
            self._original_frame_size[1],
        )

    def _show_frame(self, frame) -> None:
        super()._show_frame(frame)
        cv2.pollKey()  # for some reason Windows OS requires an additional waitKey to work properly

    @property
    def _input_parser(self):
        return WindowsInputParser()

    def _get_screen_size(self):
        user32 = ctypes.windll.user32
        screensize = 0.9 * user32.GetSystemMetrics(0), 0.9 * user32.GetSystemMetrics(1)
        return screensize

    def _get_in_focus_window_id(self):
        h_wnd = ctypes.windll.user32.GetForegroundWindow()
        pid = ctypes.wintypes.DWORD()
        ctypes.windll.user32.GetWindowThreadProcessId(h_wnd, ctypes.byref(pid))
        return pid.value

    def _get_player_window_id(self):
        user32 = ctypes.windll.user32
        user32.FindWindowW.argtypes = [LPCWSTR, LPCWSTR]
        user32.FindWindowW.restype = HWND
        h_wnd = user32.FindWindowW(None, self._window_name)
        pid = ctypes.wintypes.DWORD()
        ctypes.windll.user32.GetWindowThreadProcessId(h_wnd, ctypes.byref(pid))
        return pid.value

    def _add_default_key_functions(self) -> None:
        super()._add_default_key_functions()
        scroll_to_zoom = KeyFunction(key="mouse_scroll", func=self._set_zoom, description="zoom in and out")
        self.input_handler.register_key_function(key_function=scroll_to_zoom, callback_name="Video Control")

    def _set_zoom(self, curser_x: int, curser_y: int, scroll_direction: int):
        new_zoom_factor = max(1.0, min(5.0, self._zoom_factor + 0.5 * scroll_direction))
        if new_zoom_factor != self._zoom_factor:
            self._zoom_factor = new_zoom_factor
            adjusted_width, adjusted_height = self._screen_adjusted_frame_size
            norm_curser_position = (
                curser_x / adjusted_width,
                curser_y / adjusted_height,
            )
            w = h = 1 / self._zoom_factor
            x_min = norm_curser_position[0] * (1 - w)
            y_min = norm_curser_position[1] * (1 - h)
            self._zoom_crop_xywh = (
                int(self._original_frame_size[0] * x_min),
                int(self._original_frame_size[1] * y_min),
                int(self._original_frame_size[0] * w),
                int(self._original_frame_size[1] * h),
            )

    def crop_and_resize_frame(self, frame) -> np.ndarray:
        frame = frame[
            self._zoom_crop_xywh[1] : self._zoom_crop_xywh[1] + self._zoom_crop_xywh[3],
            self._zoom_crop_xywh[0] : self._zoom_crop_xywh[0] + self._zoom_crop_xywh[2],
        ]
        frame = cv2.resize(frame, self._screen_adjusted_frame_size)
        return frame

    def _set_icon(self):
        set_icon_windows(self._window_name, icon_path=Path(__file__).parent.parent / "icon.png")



import numpy as np

from .base_frame_edit_callback import BaseFrameEditCallback
from ..utils.video_player_utils import hist_eq, KeyFunction


class HistogramEqualizer(BaseFrameEditCallback):
    def __init__(self, enable_by_default: bool = False):
        super().__init__(enable_by_default)

    @property
    def key_function_to_register(self):
        return [
            KeyFunction(
                key="ctrl+h",
                func=self.enable_disable,
                description="Enable/Disable histogram equalization",
            ),
        ]

    def edit_frame(self, frame: np.ndarray, **kwargs) -> np.ndarray:
        if frame.dtype == "uint8":
            norm_factor = 2 ** 8 - 1
        elif frame.dtype == "uint16":
            norm_factor = 2 ** 16 - 1
        else:
            raise ValueError(f"image must be either Uint8 or Uint16 but got {frame.dtype}")

        frame = hist_eq(frame, norm_factor)
        return frame



from pathlib import Path
from typing import Union, Optional, List

from .base_video_player import VideoPlayer
from .. import FrameReader, AbstractRecorder
from ..frame_editors import BaseFrameEditCallback
from ..utils.video_player_utils import SupportedOS, CURRENT_OS

if CURRENT_OS == SupportedOS.LINUX:
    from .linux_video_player import LinuxVideoPlayer

elif CURRENT_OS == SupportedOS.WINDOWS:
    from .windows_video_player import WindowsVideoPlayer


def create_video_player(
        video_source: Union[str, Path, FrameReader],
        start_from_frame: int = 0,
        frame_edit_callbacks: Optional[List[BaseFrameEditCallback]] = None,
        record: Union[bool, AbstractRecorder] = False,
) -> VideoPlayer:
    """
    Params:
    - video_source : Union[str, Path, FrameReader] The source of the video to be played. It can be a file path, a
     directory path, or a FrameReader instance.
    - start_from_frame : int, optional The frame number to start the video from (default is 0).
    - frame_edit_callbacks : list, optional A list of frame editing callbacks (default is None).
     Each callback must be an instance of BaseFrameEditCallback.
    - record : Union[bool, AbstractRecorder], optional Whether to record the video or not (default is False).
    It can also be an instance of AbstractRecorder for custom recording functionality.
    """
    if CURRENT_OS == SupportedOS.WINDOWS:
        video_player_class = WindowsVideoPlayer

    elif CURRENT_OS == SupportedOS.LINUX:
        video_player_class = LinuxVideoPlayer

    else:
        raise ValueError(f"Unsupported OS: {CURRENT_OS}")

    return video_player_class(
        video_source=video_source,
        start_from_frame=start_from_frame,
        frame_edit_callbacks=frame_edit_callbacks,
        record=record,
    )


class DeprecatedVideoPlayer:
    def __init__(
        self,
        video_source: Union[str, Path, FrameReader],
        start_from_frame: int = 0,
        frame_edit_callbacks: Optional[List[BaseFrameEditCallback]] = None,
        record: Union[bool, AbstractRecorder] = False,
    ):
        self._video_player = create_video_player(
            video_source=video_source,
            start_from_frame=start_from_frame,
            frame_edit_callbacks=frame_edit_callbacks,
            record=record,
        )

    def run(self) -> None:
        self._video_player.run()



import subprocess
import time
from pathlib import Path

import Xlib
from Xlib import X

from ..input_management.linux_input_parser import LinuxInputParser
from ..utils.linux_os_utils import set_icon_linux
from .base_video_player import VideoPlayer


class LinuxVideoPlayer(VideoPlayer):
    def __init__(self, **video_player_kwargs):
        super().__init__(**video_player_kwargs)
        self._display = Xlib.display.Display()

    @property
    def _input_parser(self):
        return LinuxInputParser()

    def _get_in_focus_window_id(self):
        window = self._display.get_input_focus().focus
        if isinstance(window, int):
            return ""
        return window.id

    def _get_player_window_id(self):
        time.sleep(0.05)
        d = Xlib.display.Display()
        root = d.screen().root

        # Get all the window IDs on the current display
        window_ids = root.get_full_property(d.intern_atom('_NET_CLIENT_LIST'), X.AnyPropertyType).value
        for window_id in window_ids:
            # Get the window object
            window = d.create_resource_object('window', window_id)

            # Get the window name
            name = window.get_wm_name()
            if name and self._window_name in name:
                return window_id

    def _get_screen_size(self):
        screen_size_str = (
            subprocess.check_output('xrandr | grep "\*" | cut -d" " -f4', shell=True).decode().strip().split("\n")[0]
        )
        screen_w, screen_h = screen_size_str.split("x")
        screen_w = 0.85 * int(screen_w)
        screen_h = 0.85 * int(screen_h)
        return screen_w, screen_h

    def _set_icon(self):
        set_icon_linux(self._window_id, self._display, icon_path=str(Path(__file__).parent.parent / "icon.png"))



from .base_frame_edit_callback import BaseFrameEditCallback
from .base_bbox_plotter import BaseBboxPlotter
from .frame_normlizer import FrameNormalizer
from .optical_flow_plotter import OpticalFlowPlotter
from .detections_csv_plotter import DetectionsCsvPlotter
from .key_map_overlay import KeyMapOverlay
from .frame_info_overlay import FrameInfoOverlay
from .histogram_equalizer import HistogramEqualizer
from .frame_snapshot import FrameSnapshot
from .fit_frame_to_screen import FitFrameToScreen


from typing import List, TYPE_CHECKING

import numpy as np

from ..utils.video_player_utils import KeyFunction

if TYPE_CHECKING:
    from ..video_players.base_video_player import VideoPlayer


class BaseFrameEditCallback:

    def __init__(self, enable_by_default):
        self._enabled = enable_by_default

    @property
    def enabled(self):
        return self._enabled

    def setup(self, video_player: "VideoPlayer", frame) -> None:
        """
        Optionally configure more parameters according to the first incoming frame
        """

    def teardown(self) -> None:
        """
        Optionally define how the callback should close when the video player is closed
        """

    def edit_frame(
            self,
            video_player: "VideoPlayer",
            frame: np.ndarray,
            frame_num: int,
            original_frame: np.ndarray,
    ) -> np.ndarray:
        """
        This function receives the displayed frame and should return it
        after it has been altered in any way desirable by the user

        Args:
            video_player: an instance fo VideoPlayer
            frame (): the frame to be edited and displayed
            frame_num ():
            original_frame () the frame before any alterations

        Returns: the edited frame
        """
        return frame

    def enable_disable(self):
        self._enabled = not self._enabled

    @property
    def key_function_to_register(self) -> List[KeyFunction]:
        """
        Optionally return a list of KeyFunctions to be registered once the frame editor is added to the video player
        Example:
            return [
                  KeyFunction(
                      key="ctrl+f",  # A standard key combination can be any combination of modifiers with a single key
                      func=self.enable_disable,  # A function that receives no input
                      description="disable enable"  # A description of whatever this frame editor does
                  ),
                  KeyFunction(
                      key="ctrl+num", # A general num function that handles any number press
                      func=self._change_id, # this function needs to receive the number as an input
                      description="change the printed id according to the number pressed"
                  )
              ]
        """
        return []



import abc
from pathlib import Path
from queue import Empty
from typing import Optional, List, Union, Tuple
from functools import partial

import cv2
import numpy as np

from ..frame_editors import (
    BaseFrameEditCallback,
    FitFrameToScreen,
    FrameInfoOverlay,
    KeyMapOverlay,
)
from ..frame_reader import FrameReader
from ..input_management.base_input_parser import BaseInputParser
from ..input_management.input_handler import InputHandler
from ..recorder import AbstractRecorder
from ..utils.video_player_utils import (
    calc_screen_adjusted_frame_size,
    KeyFunction,
    is_window_closed_by_mouse_click,
    get_frame_reader,
    get_recorder,
)


class VideoPlayer(abc.ABC):
    def __init__(
        self,
        video_source: Union[str, Path, FrameReader],
        start_from_frame: int = 0,
        frame_edit_callbacks: Optional[List[BaseFrameEditCallback]] = None,
        record: Union[bool, AbstractRecorder] = False,
    ):
        """
        Params:
        - video_source : Union[str, Path, FrameReader] The source of the video to be played. It can be a file path, a
         directory path, or a FrameReader instance.
        - start_from_frame : int, optional The frame number to start the video from (default is 0).
        - frame_edit_callbacks : list, optional A list of frame editing callbacks (default is None).
         Each callback must be an instance of BaseFrameEditCallback.
        - record : Union[bool, AbstractRecorder], optional Whether to record the video or not (default is False).
        It can also be an instance of AbstractRecorder for custom recording functionality.
        """
        self._window_name = "CVvideoPlayer"
        self.frame_reader = get_frame_reader(video_source)
        self.input_handler = InputHandler(self._window_name)
        self._recorder = get_recorder(record)

        self._last_frame = len(self.frame_reader) - 1
        self._current_frame_num = start_from_frame

        self._screen_size = self._get_screen_size()
        self._original_frame_size = None

        self._frame_edit_callbacks = frame_edit_callbacks
        if self._frame_edit_callbacks is None:
            self._frame_edit_callbacks = [
                FitFrameToScreen(),
                FrameInfoOverlay(),
                KeyMapOverlay(),
            ]

        self._play = False
        self._exit = False
        self._add_default_key_functions()
        self._setup_callbacks()

        current_frame = self._get_current_frame()
        self._screen_adjusted_frame_size = calc_screen_adjusted_frame_size(
            screen_size=self._screen_size,
            frame_width=current_frame.shape[1],
            frame_height=current_frame.shape[0],
        )

    def __enter__(self):
        return self

    def __exit__(self, *args):
        cv2.destroyAllWindows()
        if self._recorder is not None:
            self._recorder.teardown()
        for callback in self._frame_edit_callbacks:
            callback.teardown()
        self._input_parser.stop()

    def run(self) -> None:
        try:
            self._open_player()
            self._input_parser.start()
            self._run_player_loop()
        finally:
            self.__exit__()

    def crop_and_resize_frame(self, frame) -> np.ndarray:
        frame = cv2.resize(frame, self._screen_adjusted_frame_size)
        return frame

    @abc.abstractmethod
    def _get_screen_size(self) -> Tuple[int, int]:
        """
        implemented per platform
        """

    @abc.abstractmethod
    def _get_in_focus_window_id(self) -> int:
        """
        implemented per platform
        """

    @abc.abstractmethod
    def _get_player_window_id(self) -> int:
        """
        implemented per platform
        """

    @property
    @abc.abstractmethod
    def _input_parser(self) -> BaseInputParser:
        """
        implemented per platform
        """

    @abc.abstractmethod
    def _set_icon(self):
        """
        implemented per platform
        """

    def _run_player_loop(self):
        while not self._exit:
            if self._get_in_focus_window_id() != self._window_id:
                self._input_parser.pause()
                if is_window_closed_by_mouse_click(window_name=self._window_name):
                    break
                else:
                    cv2.pollKey()
                    continue
            self._input_parser.resume()
            try:
                single_input = self._input_parser.get_input()
            except Empty:
                cv2.pollKey()
                continue

            self.input_handler.handle_input(single_input)

            if self._play:
                self._play_continuously()
            else:
                frame_for_display = self._create_frame_to_display()
                self._show_frame(frame_for_display)

    def _open_player(self) -> None:
        frame_for_display = self._create_frame_to_display()
        self._show_frame(frame_for_display)
        self._window_id = self._get_player_window_id()
        self._set_icon()
        cv2.pollKey()

    def _setup_callbacks(self):
        for callback in self._frame_edit_callbacks:
            assert isinstance(callback, BaseFrameEditCallback), (
                "frame_editor must be a derived class of" " BaseFrameEditor"
            )
            for key_function in callback.key_function_to_register:
                assert isinstance(key_function, KeyFunction), (
                    f"{callback.__class__.__name__} is trying to register a key function"
                    f" (key = {key_function.key}) which is not an instance of KeyFunction"
                )
                self.input_handler.register_key_function(
                    key_function,
                    callback.__class__.__name__,
                )

            callback.setup(video_player=self, frame=self._get_current_frame())

    def _create_frame_to_display(self) -> np.ndarray:
        original_frame = self._get_current_frame()
        frame_to_display = original_frame.copy()

        for callback in self._frame_edit_callbacks:
            if not callback.enabled:
                continue
            frame_to_display = callback.edit_frame(
                video_player=self,
                frame=frame_to_display,
                frame_num=self._current_frame_num,
                original_frame=original_frame,
            )

        if self._recorder is not None:
            self._recorder.write_frame_to_video(self, frame_to_display, self._current_frame_num)

        return frame_to_display

    def _show_frame(self, frame):
        cv2.imshow(winname=self._window_name, mat=frame)
        cv2.pollKey()

    def _play_continuously(self) -> None:
        while (not self._input_parser.has_input()) and self._play and not self._exit:
            self._change_current_frame_num(change_by=1)
            frame = self._create_frame_to_display()
            self._show_frame(frame)

    def _change_current_frame_num(self, change_by: int) -> None:
        if change_by > 0 and self._current_frame_num == self._last_frame:
            self._play = False
            return

        if change_by < 0 and self._current_frame_num == 0:
            return

        self._current_frame_num = max(0, min(self._current_frame_num + change_by, self._last_frame))

    def _get_current_frame(self) -> np.ndarray:
        return self.frame_reader.get_frame(self._current_frame_num)

    def _pause_and_change_current_frame(self, change_by: int) -> None:
        self._play = False
        self._change_current_frame_num(change_by)

    def _play_pause(self):
        self._play = not self._play

    def _set_exit_to_true(self):
        self._exit = True

    def _add_default_key_functions(self) -> None:
        default_key_functions = [
            KeyFunction("space", self._play_pause, "Play/Pause video"),
            KeyFunction("right", partial(self._pause_and_change_current_frame, 1), "Next frame"),
            KeyFunction("left", partial(self._pause_and_change_current_frame, -1), "Previous frame"),
            KeyFunction("ctrl+right", partial(self._pause_and_change_current_frame, 10), "10 frames forward"),
            KeyFunction("ctrl+left", partial(self._pause_and_change_current_frame, -10), "10 frames back"),
            KeyFunction("ctrl+shift+right", partial(self._pause_and_change_current_frame, 50), "50 frames forward"),
            KeyFunction("ctrl+shift+left", partial(self._pause_and_change_current_frame, -50), "50 frames back"),
            KeyFunction("esc", self._set_exit_to_true, "Exit gracefully"),
        ]

        for key_function in default_key_functions:
            self.input_handler.register_key_function(key_function, "Video Control")






import numpy as np
from .base_frame_edit_callback import BaseFrameEditCallback


class FitFrameToScreen(BaseFrameEditCallback):
    def __init__(
        self,
        enable_by_default: bool = True,
    ):
        super().__init__(enable_by_default)

    def edit_frame(self, video_player, frame: np.ndarray, **kwargs) -> np.ndarray:
        frame = video_player.crop_and_resize_frame(frame)
        return frame



from typing import List

from cvvideoplayer import Bbox
from cvvideoplayer.frame_editors import BaseBboxPlotter


class BboxPlottersCombiner(BaseBboxPlotter):
    def __init__(
        self,
        bbox_plotters: List[BaseBboxPlotter],
        **bboxes_plotter_kwargs,
    ):
        super().__init__(enable_by_default=True, **bboxes_plotter_kwargs)
        self._bbox_plotters = bbox_plotters

    @property
    def key_function_to_register(self):
        combined_key_dict = {}
        for bbox_plotter in self._bbox_plotters:
            combined_key_dict.update(
                {key_function.key: key_function for key_function in bbox_plotter.key_function_to_register})

        combined_key_dict.update(
            {key_function.key: key_function for key_function in super().key_function_to_register})

        return list(combined_key_dict.values())

    def edit_frame(self, frame, frame_num, original_frame, **kwargs):
        frame = super().edit_frame(frame=frame, original_frame=original_frame, frame_num=frame_num, **kwargs)
        return frame

    def get_bboxes(self, original_frame, frame_num, **kwargs) -> List[Bbox]:
        bboxes = []
        for bbox_plotter in self._bbox_plotters:
            if not bbox_plotter.enabled:
                continue
            bboxes.extend(bbox_plotter.get_bboxes(
                original_frame=original_frame,
                frame_num=frame_num,
                **kwargs
            ))
        return bboxes

    def teardown(self):
        for bbox_plotter in self._bbox_plotters:
            bbox_plotter.teardown()
